/*
A menu that should be close to the user, it's the user's status,
their friends.  All about them.  It's a user-focused-menu.

Copyright 2009 Canonical Ltd.

Authors:
    Ted Gould <ted@canonical.com>

This program is free software: you can redistribute it and/or modify it 
under the terms of the GNU General Public License version 3, as published 
by the Free Software Foundation.

This program is distributed in the hope that it will be useful, but 
WITHOUT ANY WARRANTY; without even the implied warranties of 
MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR 
PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along 
with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <glib.h>
#include <glib-object.h>
#include <gtk/gtk.h>
#include <libdbusmenu-glib/menuitem.h>
#include <libdbusmenu-gtk/menu.h>

#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-bindings.h>

#include <libindicator/indicator.h>
#include <libindicator/indicator-object.h>
#include <libindicator/indicator-service-manager.h>
#include <libido/idoentrymenuitem.h>

#include "dbus-shared-names.h"
#include "me-service-client.h"

#define INDICATOR_ME_TYPE            (indicator_me_get_type ())
#define INDICATOR_ME(obj)            (G_TYPE_CHECK_INSTANCE_CAST ((obj), INDICATOR_ME_TYPE, IndicatorMe))
#define INDICATOR_ME_CLASS(klass)    (G_TYPE_CHECK_CLASS_CAST ((klass), INDICATOR_ME_TYPE, IndicatorMeClass))
#define IS_INDICATOR_ME(obj)         (G_TYPE_CHECK_INSTANCE_TYPE ((obj), INDICATOR_ME_TYPE))
#define IS_INDICATOR_ME_CLASS(klass) (G_TYPE_CHECK_CLASS_TYPE ((klass), INDICATOR_ME_TYPE))
#define INDICATOR_ME_GET_CLASS(obj)  (G_TYPE_INSTANCE_GET_CLASS ((obj), INDICATOR_ME_TYPE, IndicatorMeClass))

#define DEFAULT_ICON   "user-offline"

typedef struct _IndicatorMe      IndicatorMe;
typedef struct _IndicatorMeClass IndicatorMeClass;

struct _IndicatorMeClass {
	IndicatorObjectClass parent_class;
};

struct _IndicatorMe {
	IndicatorObject parent;
	IndicatorServiceManager * service;
};

GType indicator_me_get_type (void);

/* Indicator stuff */
INDICATOR_SET_VERSION
INDICATOR_SET_TYPE(INDICATOR_ME_TYPE)

/* Globals */
static GtkImage * status_image = NULL;
static DBusGProxy * status_proxy = NULL;

/* Prototypes */
static GtkLabel * get_label (IndicatorObject * io);
static GtkImage * get_icon (IndicatorObject * io);
static GtkMenu * get_menu (IndicatorObject * io);

static void indicator_me_class_init (IndicatorMeClass *klass);
static void indicator_me_init       (IndicatorMe *self);
static void indicator_me_dispose    (GObject *object);
static void indicator_me_finalize   (GObject *object);
static void connection_changed      (IndicatorServiceManager * sm, gboolean connected, gpointer userdata);
static void status_icon_cb (DBusGProxy * proxy, char * icons, GError *error, gpointer userdata);

G_DEFINE_TYPE (IndicatorMe, indicator_me, INDICATOR_OBJECT_TYPE);

static void
indicator_me_class_init (IndicatorMeClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);

	object_class->dispose = indicator_me_dispose;
	object_class->finalize = indicator_me_finalize;

	IndicatorObjectClass * io_class = INDICATOR_OBJECT_CLASS(klass);
	io_class->get_label = get_label;
	io_class->get_image = get_icon;
	io_class->get_menu = get_menu;

	return;
}

static void
indicator_me_init (IndicatorMe *self)
{

	/* Init variables */
	self->service = NULL;

	/* Do stuff with them */
	self->service = indicator_service_manager_new_version(INDICATOR_ME_DBUS_NAME, INDICATOR_ME_DBUS_VERSION);
	g_signal_connect(G_OBJECT(self->service), INDICATOR_SERVICE_MANAGER_SIGNAL_CONNECTION_CHANGE, G_CALLBACK(connection_changed), self);

	return;
}

static void
indicator_me_dispose (GObject *object)
{

	G_OBJECT_CLASS (indicator_me_parent_class)->dispose (object);
	return;
}

static void
indicator_me_finalize (GObject *object)
{

	G_OBJECT_CLASS (indicator_me_parent_class)->finalize (object);
	return;
}

static GtkLabel *
get_label (IndicatorObject * io)
{
	GtkLabel * returnval = GTK_LABEL(gtk_label_new(g_get_user_name()));
	gtk_widget_show(GTK_WIDGET(returnval));
	return returnval;
}

static GtkImage *
get_icon (IndicatorObject * io)
{
	if (status_image == NULL) {
		/* Will create the status icon if it doesn't exist already */
		status_icon_cb(NULL, DEFAULT_ICON, NULL, NULL);
	}
	return status_image;
}

static void
status_icon_cb (DBusGProxy * proxy, char * icons, GError *error, gpointer userdata)
{
	g_return_if_fail(icons != NULL);
	g_return_if_fail(icons[0] != '\0');

	g_debug("Changing status icon: '%s'", icons);

	if (status_image == NULL) {
		status_image = GTK_IMAGE(gtk_image_new_from_icon_name(DEFAULT_ICON, GTK_ICON_SIZE_MENU));
		gtk_widget_show(GTK_WIDGET(status_image));
	}

	gtk_image_set_from_icon_name(status_image, icons, GTK_ICON_SIZE_MENU);

	return;
}

static void
status_icon_changed (DBusGProxy * proxy, gchar * icon, gpointer userdata)
{
	return status_icon_cb(proxy, icon, NULL, NULL);
}

static void
connection_changed (IndicatorServiceManager * sm, gboolean connected, gpointer userdata)
{
	if (connected) {
		if (status_proxy == NULL) {
			GError * error = NULL;

			DBusGConnection * sbus = dbus_g_bus_get(DBUS_BUS_SESSION, NULL);

			status_proxy = dbus_g_proxy_new_for_name_owner(sbus,
														   INDICATOR_ME_DBUS_NAME,
														   INDICATOR_ME_SERVICE_DBUS_OBJECT,
														   INDICATOR_ME_SERVICE_DBUS_INTERFACE,
														   &error);

			if (error != NULL) {
				g_warning("Unable to get status proxy: %s", error->message);
				g_error_free(error);
			}

			dbus_g_proxy_add_signal(status_proxy, "StatusIconsChanged", G_TYPE_STRING, G_TYPE_INVALID);
			dbus_g_proxy_connect_signal(status_proxy, "StatusIconsChanged", G_CALLBACK(status_icon_changed), NULL, NULL);
		}

		org_ayatana_indicator_me_service_status_icons_async(status_proxy, status_icon_cb, NULL);
	} else {
		/* If we're disconnecting, go back to offline */
		status_icon_cb(NULL, DEFAULT_ICON, NULL, NULL);
	}

	return;
}

static void
entry_prop_change_cb (DbusmenuMenuitem *mi, gchar *prop, GValue *value, GtkEntry *entry)
{
	if (g_strcmp0 (prop, ENTRY_MENUITEM_PROP_TEXT) == 0) {
		gtk_entry_set_text (entry, g_value_get_string (value));
	}
}

static void
activate_cb (GtkEntry *entry, DbusmenuMenuitem *gmi)
{
	GValue value = { 0 };
	g_value_init (&value, G_TYPE_STRING);
	g_value_set_static_string (&value, gtk_entry_get_text (entry));

	g_debug ("user typed: %s", g_value_get_string (&value));

	dbusmenu_menuitem_handle_event (gmi, "send", &value, gtk_get_current_event_time());
}

static gboolean
new_entry_item (DbusmenuMenuitem * newitem,
				DbusmenuMenuitem * parent,
				DbusmenuClient * client)
{
	g_return_val_if_fail(DBUSMENU_IS_MENUITEM(newitem), FALSE);
	g_return_val_if_fail(DBUSMENU_IS_GTKCLIENT(client), FALSE);
	/* Note: not checking parent, it's reasonable for it to be NULL */

	IdoEntryMenuItem *ido = IDO_ENTRY_MENU_ITEM (ido_entry_menu_item_new ());
	GtkEntry *entry = GTK_ENTRY(ido_entry_menu_item_get_entry (ido));
	if (dbusmenu_menuitem_property_get (newitem, ENTRY_MENUITEM_PROP_TEXT) != NULL)
		gtk_entry_set_text(entry, dbusmenu_menuitem_property_get(newitem, ENTRY_MENUITEM_PROP_TEXT));

	dbusmenu_gtkclient_newitem_base(DBUSMENU_GTKCLIENT(client), newitem, GTK_MENU_ITEM(ido), parent);
	/* disconnect the activate signal that newitem_base connected with the wrong
	   widget, ie menuitem, and re-connect it with the /entry/ instead */
	gulong signal_id = g_signal_handler_find (GTK_MENU_ITEM (ido), G_SIGNAL_MATCH_DATA, 0, 0, 0, 0, parent);
	g_signal_handler_disconnect(GTK_MENU_ITEM (ido), signal_id);

	g_signal_connect (DBUSMENU_MENUITEM (newitem), DBUSMENU_MENUITEM_SIGNAL_PROPERTY_CHANGED, G_CALLBACK (entry_prop_change_cb), entry);
	g_signal_connect (GTK_ENTRY (entry), "activate", G_CALLBACK (activate_cb), newitem);

	return TRUE;
}

/* Builds the dbusmenu for the service. */
static GtkMenu *
get_menu (IndicatorObject * io)
{
	DbusmenuGtkMenu *menu = dbusmenu_gtkmenu_new(INDICATOR_ME_DBUS_NAME, INDICATOR_ME_DBUS_OBJECT);
	DbusmenuGtkClient * client = dbusmenu_gtkmenu_get_client(menu);

	dbusmenu_client_add_type_handler(DBUSMENU_CLIENT(client), ENTRY_MENUITEM_TYPE, new_entry_item);

	return GTK_MENU (menu);
}
